/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * Code for Algorithm 4.7 (TDR-c-optimal), p. 102  
 *
 * this file: algo04_07.c
 *
 * For a demo how to use this algorithm see the file "demo_algo04_07.c".
 *
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "uniform.h"

#define T(c,x) exp((c)*log(x))
#define DT(c,x) ((c)*exp(((c)-1.)*log(x)))
#define FT(c,x) (exp(log(x)*(1.+1./(c)))/(1.+1./(c)))
#define FTI(c,x) (exp(log((x)*((c)+1.)/(c))*((c)/(1.+(c)))))
#define TI(c,x) exp(log(x)/(c))


#define SQR(a) ((a)*(a))
 static double *p=NULL,*c=NULL, *tfp=NULL, *dtfp=NULL, *s=NULL, *a=NULL, 
       atotal, (*pdfh)(double);
 static int NN = 0;


int setup04_07(double (*pdf)(double), 
               double (*dpdf)(double),  double (*ddpdf)(double),  
               int N, double constructionp[])
/*
 * Algorithm 4.7 (TDR-c-optimal), p. 102  
 *
 * pdf... function pointer to the density
 * dpdf... function pointer to the derivative of the density
 * ddpdf...function pointer to the second derivative of the density
 * N  ... number of construction points
 * constructionp[] ... array holding the construction points
 * constructionp[0] is the left, constructionp[N-1] the right border of the domain
 *
 */
{    
     int i;  

     if(p!=NULL){
       free(p);
       free(c);
       free(tfp);
       free(dtfp);
       free(s);
       free(a);
     }
     NN = N;
     pdfh =pdf;
     p = malloc((N+1)*sizeof(double));  
     c = malloc((N+1)*sizeof(double));  
     tfp = malloc((N+1)*sizeof(double));  
     dtfp = malloc((N+1)*sizeof(double));  
     s = malloc((N+1)*sizeof(double));  
     a = malloc((N+1)*sizeof(double));  
     for(i=1;i<=N;i++) p[i] = constructionp[i-1];
/*1*/for(i=2;i<=N;i++) c[i]=1.-(*ddpdf)(p[i])*(*pdf)(p[i])/SQR((*dpdf)(p[i]));
/*2*/ for(i=1;i<=N;i++){
       tfp[i] = T(c[i],(*pdf)(p[i]));
       dtfp[i] = DT(c[i],(*pdf)(p[i]))*(*dpdf)(p[i]);
     }
/*3*/for(i=2;i<=N;i++) s[i]=(tfp[i]-T(c[i],(*pdf)(p[i-1])))/(p[i]-p[i-1]);
     atotal=0.;
     for(i=2;i<=N;i++){
/*4*/  a[i] = (FT(c[i],tfp[i])-FT(c[i],tfp[i]+dtfp[i]*(p[i-1]-p[i])))/dtfp[i];
/*5*/  atotal += a[i];
     }
     for(i=1;i<=N;i++){
       printf("i%d <%f> area %f c %f\n",i,p[i],a[i],c[i]);
       printf("tfp %f dtfp %f s %f\n",tfp[i],dtfp[i],s[i]);
     }
     printf("total area below hat: %f\n",atotal);
     return 1;
}


double generator04_07()
/* is called after the setup to generator reandom variates */
{     double u, v, asum, sx, hx, x;
      int j;  

      if(p==NULL){
	printf("ERROR generator04_07: you have to run the steup first!!\n");
	return -9999.;
      }
/*6*/ while(1){
/*7*/   v = atotal*uniform();
        j = 1;
	asum = a[1];
	while(v > asum){
	  j++;
	  asum += a[j];
	}
/*8*/   v = v - (asum-a[j]);
/*9*/   x = p[j]+(FTI(c[j],dtfp[j]*v+FT(c[j],tfp[j]+dtfp[j]*(p[j-1]-p[j])))-tfp[j])/dtfp[j];
/*10*/  hx = TI(c[j],tfp[j]+dtfp[j]*(x-p[j]));
/*11*/  u = uniform();
/*12*/  sx = TI(c[j],tfp[j]+s[j]*(x-p[j]));
/*13*/  if(u*hx <= sx) 
/*14*/    return x;
/*15*/  if(u*hx <= (*pdfh)(x))
/*16*/    return x;
      }
}

int free04_07()
/* frees the allocated memory */
{
      free(p);
      p = NULL;
      free(c);
      c = NULL;
      free(tfp);
      tfp = NULL;
      free(dtfp);
      dtfp = NULL;
      free(s);
      s = NULL;
      free(a);
      a = NULL;
      return 1;
}


